/* 
   conn.c - connections handling

   Copyright (C) Szymon Acedanski 2003-2004
*/

/*
    ggrelay - Gadu-Gadu instant messager relaying agent
    Copyright (C) 2003-2004 Szymon Acedanski

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "config.h"
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <errno.h>

#ifdef HAVE_LINUX_NETFILTER_IPV4_H
#include <limits.h>
#include <linux/netfilter_ipv4.h>
#endif

#if HAVE_POLL
#if HAVE_POLL_H
#include <poll.h>
#else
#undef HAVE_POLL
#endif
#endif

#include "ggrelay.h"

static pthread_t monitor_thread;

static inline struct gg_server_desc* next_server(struct gg_server_desc* s)
{
	return s->next?s->next:gg_servers;
}

static struct gg_server_desc* select_server(void)
{
	struct gg_server_desc* next = gg_servers_iter;

	do
	{
		next = next_server(next);
		next->skipcnt++;
	} while (next->skipcnt <= next->errs);
	next->skipcnt = 0;

	return gg_servers_iter = next;
}

#ifdef SO_ORIGINAL_DST
static int verify_server(struct sockaddr_in* sin)
{
	struct gg_server_desc* s = gg_servers;
	pthread_mutex_lock(&gg_servers_mutex);
	while (s)
	{
		if (sin->sin_addr.s_addr == s->addr.sin_addr.s_addr &&
				sin->sin_port == s->addr.sin_port)
		{
			pthread_mutex_unlock(&gg_servers_mutex);
			return 0;
		}
		s = s->next;
	}
	pthread_mutex_unlock(&gg_servers_mutex);
	return 1;
}
#endif

static void register_server_error(struct gg_server_desc* s)
{
	if (s == NULL) return;
	if (s->errs >= GG_MAX_SERVER_ERRS) return;
	s->errs++;
}

static void register_server_ok(struct gg_server_desc* s)
{
	if (s == NULL) return;
	s->errs = 0;
}

static void block_signals(void)
{
	sigset_t s;
	sigfillset(&s);
	pthread_sigmask(SIG_BLOCK, &s, NULL);
}

static void gg_cleanup_fd(struct gg_fd* f)
{
	if (f->state == GG_S_DESTROY) return;
	gg_condlog("closed fd %d", f->fd);
#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
	{
		gg_ssl_shutdown(f, SHUT_RDWR);
		gg_ssl_close(f);
	}
	else
#endif
	{
		shutdown(f->fd, SHUT_RDWR);
		gg_close_and_unmap_fd(f);
	}
	gg_masq_close_handler(f);
	f->state = GG_S_DESTROY;
	gg_assert(!list_empty(&f->list));
	if (f->type == GG_T_CHATCOUNTED)
		gg_max_chat_conns++;
	else if (f->type == GG_T_DCCCOUNTED)
		gg_max_dcc_conns++;
	if (f->expby)
	{
		gg_cleanup_fd(f->expby);
		f->expby = NULL;
	}
	if (f->parent)
	{
		gg_put_fd(f->parent);
		f->parent = NULL;
	}
	if (f->cexp)
	{
		list_del(&f->cexp->list);
		free(f->cexp);
		f->cexp = NULL;
	}
}

void gg_close_fd(struct gg_fd* f)
{
	DEBUGP("closing fd %d", f->fd);
	if (f->twin && f->twin != f)
	{
		if (f->readoffset)
		{
#ifdef HAVE_OPENSSL
			if (f->twin->flags & GG_FLAG_SSL)
				gg_ssl_shutdown(f->twin, SHUT_RD);
			else
#endif
				shutdown(f->twin->fd, SHUT_RD);
			f->twin->state = GG_S_CLOSING;
			f->twin->twin = f->twin;
			f->twin->readoffset = f->readoffset;
			f->twin->writeoffset = f->readoffset;
			/* must not memcpy buffers, since they can have different sizes */
			free(f->twin->buffer);
			f->twin->buffer = f->buffer;
			f->buffer = NULL;
			f->twin->bufsize = f->bufsize;
			f->bufsize = -1;
		}
		else
			gg_cleanup_fd(f->twin);
	}
	gg_cleanup_fd(f);
}

void gg_abort_fd(struct gg_fd* f)
{
	if (f->state == GG_S_DESTROY) return;
	gg_condlog("aborting fd %d", f->fd);
	if (f->twin && f->twin != f)
		gg_cleanup_fd(f->twin);
	gg_cleanup_fd(f);
}

static inline void gg_adjust_sockbuffs(int fd)
{
	int foo = GG_SOCK_BUFSIZE;
	setsockopt(fd, SOL_SOCKET, SO_SNDBUF, &foo, sizeof(int));
	setsockopt(fd, SOL_SOCKET, SO_RCVBUF, &foo, sizeof(int));
}

#if HAVE_POLL

int gg_my_connect(int sockfd, const struct sockaddr* serv_addr,
		socklen_t addrlen)
{
	int foo = 1;
	int foolen = sizeof(foo);
	int ret = -1;
	
	if (ioctl(sockfd, FIONBIO, &foo, sizeof(foo)))
		DEBUGP("ioctl(FIONBIO) failed with '%s'", strerror(errno));
	
	if (connect(sockfd, serv_addr, addrlen))
	{
		if (errno == EINPROGRESS)
		{
			struct pollfd pfd;
			int n;
			int sockerr;
			
			DEBUGP("polling() for connection completion");
			pfd.fd = sockfd;
			pfd.events = POLLOUT|POLLERR|POLLHUP;

			do
			{
				n = poll(&pfd, 1, gg_connect_timeout);
			} while (n == -1 && errno != EINTR);

			if (n == -1)
			{
				DEBUGP("poll() failed with '%s' (fd=%d)",
						strerror(errno), sockfd);
				goto fin;
			}

			if (n == 0)
			{
				DEBUGP("connection timed out (fd=%d)", sockfd);
				errno = ETIMEDOUT;
				goto fin;
			}

			/*if (pfd.revents & POLLERR)
			{
				DEBUGP("poll() reported error (fd=%d)", sockfd);
				errno = ECONNREFUSED;
				goto fin;
			}*/

			sockerr = getsockopt(sockfd, SOL_SOCKET, SO_ERROR,
					&foo, &foolen);
			if (foo)
			{
				DEBUGP("error connecting (fd=%d, err=%d)", sockfd, foo);
				errno = foo;
				goto fin;
			}
		}
		else
		{
			DEBUGP("connect() failed with '%s' (fd=%d)",
					strerror(errno), sockfd);
			goto fin;
		}
	}

	ret = 0;
	DEBUGP("successfully established (fd=%d)", sockfd);
	
fin:
	foo = 0;
	if (ioctl(sockfd, FIONBIO, &foo, sizeof(foo)))
		DEBUGP("ioctl(!FIONBIO) failed with '%s'", strerror(errno));
	return ret;
}

#else /* HAVE_POLL */

int gg_my_connect(int sockfd, const struct sockaddr* serv_addr,
		socklen_t addrlen)
{
	return connect(sockfd, serv_addr, addrlen);
}

#endif /* HAVE_POLL */

static void* gg_accept_chat(void* __info)
{
	struct gg_connect_info* _info = (struct gg_connect_info*)__info;
	struct gg_connect_info info = *_info;
	struct gg_server_desc* svr = NULL;
	struct sockaddr_in sin;
	struct sockaddr_in sin_svr;
	int fd = -1;
	int foo = 1;
	int retry = 0;
	struct gg_fd* f;
	uint32_t the_outer_ip = gg_outer_ip;
#ifdef SO_ORIGINAL_DST
	socklen_t sinlen = sizeof(sin_svr);
#endif

	free(__info);
	block_signals();
	
#ifdef HAVE_OPENSSL
	if (info.twin->flags & GG_FLAG_SSL)
		if (gg_ssl_accept_connection(info.twin))
			goto failure;
#endif

	if (ioctl(info.twin->fd, FIONBIO, &foo, sizeof(foo)))
	{
		DEBUGP("ioctl(FIONBIO) failed with '%s'", strerror(errno));
		goto failure;
	}
	
	if ((fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
	{
		DEBUGP("socket() failed with '%s'", strerror(errno));
		goto failure;
	}

	gg_adjust_sockbuffs(fd);

	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	if (the_outer_ip)
		sin.sin_addr.s_addr = the_outer_ip;
	else
	{
		DEBUGP("warning: no external ip known, binding to whatever");
		sin.sin_addr.s_addr = INADDR_ANY;
	}
	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		DEBUGP("bind() failed with '%s'", strerror(errno));
		goto closefd;
	}

#ifdef SO_ORIGINAL_DST
	if (!getsockopt(info.twin->fd, SOL_IP, SO_ORIGINAL_DST, &sin_svr, &sinlen))
	{
		DEBUGP("getsockopt(SO_ORIGINAL_DST) succeeded");

		/* never connect to unlisted server; if !gg_drop_unlisted then fall back
		   to listed one. */
		
		if (verify_server(&sin_svr))
		{
			gg_condlog("client tried to connect to an unlisted server %u.%u.%u.%u",
				NIPQUAD(sin_svr.sin_addr.s_addr));
			if (gg_drop_unlisted)
				goto failure;
		}
		else
			goto svr_selected;
	}
#endif

svr_select_from_list:
	pthread_mutex_lock(&gg_servers_mutex);
	svr = select_server();
	memcpy(&sin_svr, &svr->addr, sizeof(sin_svr));
	pthread_mutex_unlock(&gg_servers_mutex);

#ifdef SO_ORIGINAL_DST
svr_selected:
#endif

	DEBUGP("connecting to %u.%u.%u.%u:%hd (retry=%d)", NIPQUAD(sin_svr.sin_addr.s_addr),
			ntohs(sin_svr.sin_port), retry);
	if (gg_my_connect(fd, (struct sockaddr*)&sin_svr, sizeof(sin_svr)))
	{
		DEBUGP("connect() failed with '%s'", strerror(errno));
		register_server_error(svr);
		if (++retry >= gg_retry_count)
			goto closefd;
		else
			goto svr_select_from_list;
	}

	register_server_ok(svr);

	if (!(f = gg_alloc_fd(fd, gg_dynamic_buffers ?
					GG_CHATREPEATER_BUFSIZE_DYNAMIC :
					GG_CHATREPEATER_BUFSIZE_STATIC)))
		goto closefd;
	f->type = GG_T_NOTCOUNTED;
	f->twin = info.twin;
	f->fd = fd;
	info.twin->twin = f;
	gg_masq_init_chat_req(info.twin);
	gg_masq_init_chat_resp(f);
	
	gg_condlog("established chat connection on fds %d:%d",
			f->twin->fd, f->fd);
	
	pthread_mutex_lock(&gg_fds_mutex);
	if (gg_outer_ip != the_outer_ip) /* must have changed in the meantime */
	{
		pthread_mutex_unlock(&gg_fds_mutex);
		DEBUGP("outer IP changed while servicing connection request");
		gg_close_and_unmap_fd(f);
		goto failure;
	}
	__gg_attach_fd(f);
	__gg_attach_fd(info.twin);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_kick_main_thread();

	return (void*)0;
	
closefd:
	close(fd);
failure:
	gg_condlog("failure establishing chat connection on %d:%d",
			info.twin->fd, fd);
	pthread_mutex_lock(&gg_fds_mutex);
	gg_close_and_unmap_fd(info.twin);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_put_fd(info.twin); /* not attached, must be put explicitly */
	return (void*)1;
}

void* gg_establish_dcc(void* __info)
{
	struct gg_connect_info* _info = (struct gg_connect_info*)__info;
	struct gg_connect_info info = *_info;
	struct sockaddr_in sin;
	int fd = -1;
	struct gg_fd* f;
	free(__info);
	
	block_signals();

	DEBUGP("dcc helper thread starting");
	
	if (!gg_outer_ip)
	{
		DEBUGP("no external ip, dropping connection");
		goto failure;
	}

	if ((fd = socket(PF_INET, SOCK_STREAM, 0)) == -1)
		goto failure;

	gg_adjust_sockbuffs(fd);
	sin.sin_family = AF_INET;
	sin.sin_port = 0;
	sin.sin_addr.s_addr = gg_outer_ip;
	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		DEBUGP("bind() failed with '%s'", strerror(errno));
		goto closefd;
	}

	sin.sin_port = htons(info.nat_port);
	sin.sin_addr.s_addr = info.nat_ip;
	if (gg_my_connect(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		DEBUGP("connect() failed with '%s'", strerror(errno));
		goto closefd;
	}

	if (!(f = gg_alloc_fd(fd, GG_DCCREPEATER_BUFSIZE)))
		goto closefd;
	f->type = GG_T_NOTCOUNTED;
	f->state = GG_S_REPEATMODE;
	f->twin = info.twin;
	f->fd = fd;

	gg_condlog("established dcc connection on fds %d:%d",
			f->twin->fd, f->fd);

	pthread_mutex_lock(&gg_fds_mutex);
	info.twin->twin = f;
	if (info.twin->state == GG_S_DESTROY)
	{
		DEBUGP("caught just-destroyed condition on %d:%d",
				info.twin->fd, fd);
		gg_put_fd(f);
		gg_put_fd(info.twin);
		pthread_mutex_unlock(&gg_fds_mutex);
		close(fd);
		return (void*)1;
	}
	info.twin->state = GG_S_REPEATMODE;
	__gg_attach_fd(f);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_kick_main_thread();
	gg_put_fd(info.twin);

	return (void*)0;
	
closefd:
	close(fd);
failure:
	gg_condlog("failure establishing dcc connection on %d:%d",
			info.twin->fd, fd);
	pthread_mutex_lock(&gg_fds_mutex);
	gg_cleanup_fd(info.twin);
	pthread_mutex_unlock(&gg_fds_mutex);
	gg_put_fd(info.twin);
	return (void*)1;
}

int gg_accept_chat_connection(int lfd, int ssl)
{
	struct sockaddr_in sin;
	socklen_t addrlen = sizeof(sin);
	int fd = accept(lfd, (struct sockaddr*)&sin, &addrlen);
	struct gg_fd* f;
	struct gg_connect_info* info;
	pthread_attr_t attr;
	pthread_t dummy;

	/* gg_fds_mutex is locked. beware. */

	if (fd == -1)
		return -1;

	if (!gg_max_chat_conns)
	{
		gg_condlog("REJECTED chat connection fd %d from %u.%u.%u.%u:%u",
				fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
		goto closefd;
	}

	gg_condlog("accepted chat connection fd %d from %u.%u.%u.%u:%u",
			fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
	
	if (!(f = gg_alloc_fd(fd, gg_dynamic_buffers ?
					GG_CHATREPEATER_BUFSIZE_DYNAMIC :
					GG_CHATREPEATER_BUFSIZE_STATIC)))
		goto closefd;
	gg_adjust_sockbuffs(fd);
	f->ip = sin.sin_addr.s_addr;
	if (!(info = (struct gg_connect_info*)malloc(sizeof(*info))))
		goto freef;
	f->type = GG_T_CHATCOUNTED;
	f->flags |= GG_FLAG_CLIENTCONN;
	if (ssl)
		f->flags |= GG_FLAG_SSL;
	info->twin = f;
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_DETACHED);
	DEBUGP("spawning helper thread...");
	if (pthread_create(&dummy, &attr, gg_accept_chat, info))
		goto freeinfo;
	gg_max_chat_conns--;
	return 0;
	
freeinfo:
	free(info);
freef:
	gg_put_fd(f);
closefd:
	close(fd);
	return -1;
}

int gg_accept_dcc_connection(void)
{
	struct sockaddr_in sin;
	socklen_t addrlen = sizeof(sin);
	int fd = accept(gg_dcc_listener, (struct sockaddr*)&sin, &addrlen);
	struct gg_fd* f;

	/* gg_fds_mutex is locked. beware. */
	
	if (fd == -1)
		return -1;

	if (!gg_max_dcc_conns)
	{
		gg_condlog("REJECTED dcc connection fd %d from %u.%u.%u.%u:%u",
				fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
		goto closefd;
	}

	gg_condlog("accepted dcc connection fd %d from %u.%u.%u.%u:%u",
			fd, NIPQUAD(sin.sin_addr), ntohs(sin.sin_port));
	
	if (!(f = gg_alloc_fd(fd, GG_DCCREPEATER_BUFSIZE)))
		goto closefd;
	gg_adjust_sockbuffs(fd);
	f->ip = sin.sin_addr.s_addr;
	f->type = GG_T_DCCCOUNTED;
	gg_masq_init_dcc_req(f);
	__gg_attach_fd(f);
	gg_max_dcc_conns--;
	return 0;
	
closefd:
	close(fd);
	return -1;
}

static void* gg_ip_monitor_thread(void* dummy)
{
	DEBUGP("IP monitor thread starting");
	for (;;)
	{
		gg_update_outer_ip();
		sleep(5);
	}
}

int gg_setup_ip_monitor(void)
{
	pthread_attr_t attr;
	pthread_attr_init(&attr);
	DEBUGP("spawning IP monitor thread...");
	return pthread_create(&monitor_thread, &attr, gg_ip_monitor_thread, NULL);
}

void gg_kill_ip_monitor(void)
{
	DEBUGP("cancelling IP monitor thread ...");
	pthread_cancel(monitor_thread);
	DEBUGP("done.");
}

void gg_reset_all_fds(void)
{
	struct gg_fd* i;
	if (list_empty(&gg_fds)) return;
	gg_condlog("resetting all connections");
	list_for_each_entry(i, &gg_fds, list)
		gg_abort_fd(i);
	gg_kick_main_thread();
}

void gg_update_outer_ip(void)
{
	uint32_t newip;
	
	gg_assert(gg_outer_iface != NULL);
	newip = gg_ifname_to_ip(gg_outer_iface);
	DEBUGP("iface=%s oldip=%u.%u.%u.%u newip=%u.%u.%u.%u", gg_outer_iface,
			NIPQUAD(gg_outer_ip), NIPQUAD(newip));
	if (newip == gg_outer_ip)
		return;

	pthread_mutex_lock(&gg_fds_mutex);
	if (newip)
	{
		gg_condlog("(re)binding dcc listener to %u.%u.%u.%u", NIPQUAD(newip));
		if (gg_dcc_listener >= 0)
			close(gg_dcc_listener);
		if ((gg_dcc_listener = gg_setup_listener(newip, gg_dcc_port)) < 0)
		{
			gg_warn("error '%s' while binding DCC listener socket to %u.%u.%u.%u:%hd",
					strerror(errno), NIPQUAD(newip), gg_dcc_port);
			newip = 0;
		}
	}
	else
		gg_condlog("lost current IP, waiting for another");
	if (gg_outer_ip != newip)
	{
		gg_outer_ip = newip;
		if (newip)
			gg_reset_all_fds();
	}
	pthread_mutex_unlock(&gg_fds_mutex);
}

int gg_setup_listener(uint32_t ip, uint16_t port)
{
	int fd = socket(PF_INET, SOCK_STREAM, 0);
	struct sockaddr_in sin;
	int dummy = 1;

	gg_condlog("setting up listener on %u.%u.%u.%u:%u", NIPQUAD(ip), port);
	
	if (fd == -1)
	{
		gg_error("socket() failed with '%s'", strerror(errno));
		return -1;
	}
	
	sin.sin_family = AF_INET;
	sin.sin_port = htons(port);
	sin.sin_addr.s_addr = ip;

	if (setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (char*)&dummy, sizeof(dummy)))
	{
		gg_error("setsockopt(SO_REUSEADDR) failed with '%s'", strerror(errno));
		goto closesock;
	}

	if (bind(fd, (struct sockaddr*)&sin, sizeof(sin)))
	{
		gg_error("bind() failed with '%s'", strerror(errno));
		goto closesock;
	}

	if (listen(fd, GG_LISTENER_BACKLOG))
	{
		gg_error("listen() failed with '%s'", strerror(errno));
		goto closesock;
	}
	return fd;
	
closesock:
	close(fd);
	return -1;
}

int gg_fd_wants_reading(struct gg_fd* f)
{
	if (f->flags & GG_FLAG_READFORWRITING) return 1;
	if (f->readoffset == f->bufsize) return 0;
	if (f->state == GG_S_CLOSING) return 0;
	if (f->state == GG_S_WRTCMPL) return 0;
	if (f->state == GG_S_SUSPENDED) return 0; /* is it necessary? */
	if (f->twin && f->twin->flags & GG_FLAG_WAITFORME) return 0;
	return 1;
}

int gg_fd_wants_writing(struct gg_fd* f)
{
	if (f->flags & GG_FLAG_WRITEFORREADING) return 1;
	if (f->state == GG_S_SUSPENDED) return 0; /* is it necessary? */
	if (!f->twin) return 0;
	if (f->twin->writeoffset == 0) return 0;
	if (f->twin->flags & GG_FLAG_WAITFORME) return 0;
	return 1;
}

static inline void gg_check_expected(struct gg_fd* f)
{
	while (f->state == GG_S_EXPECTING && f->readoffset >= f->expectlen
			&& f->writeoffset == 0)
		f->expectfn(f);
}

void gg_fd_handle_reading(struct gg_fd* f)
{
	int size;

	if (f->state == GG_S_DESTROY) return;
	if (!gg_fd_wants_reading(f)) return;

#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
		size = gg_ssl_read(f);
	else
#endif
		size = read(f->fd, f->buffer + f->readoffset,
				f->bufsize - f->readoffset);

	DEBUGP("fd=%d %s twin=%d read=%d ro=%d wo=%d state=%d", f->fd,
			(f->flags & GG_FLAG_SSL) ? "(ssl)" : "", 
			f->twin?f->twin->fd:-1, size, f->readoffset, f->writeoffset, f->state);

	if (size == -2)
		return;
	else if (size <= 0) /* aborting on error does not produce anything good */
		gg_close_fd(f);
	else
	{
		f->readoffset += size;
		if (f->state == GG_S_EXPECTING)
			gg_check_expected(f);
		else
			f->writeoffset = f->readoffset;
	}
}

void gg_fd_handle_writing(struct gg_fd* f)
{
	int written;

	if (f->state == GG_S_DESTROY) return;
	if (!f->twin)
	{
		gg_abort_fd(f);
		return;
	}

	if (f->state == GG_S_CLOSING && f->writeoffset == 0)
	{
		gg_close_fd(f);
		return;
	}

	if (!gg_fd_wants_writing(f))
		return;
	
#ifdef HAVE_OPENSSL
	if (f->flags & GG_FLAG_SSL)
		written = gg_ssl_write(f);
	else
#endif
		written = write(f->fd, f->twin->buffer, f->twin->writeoffset);

	DEBUGP("fd=%d %s twin=%d written=%d ro=%d wo=%d state=%d", f->fd, 
			(f->flags & GG_FLAG_SSL) ? "(ssl)" : "", f->twin->fd, written,
			f->twin->readoffset, f->twin->writeoffset, f->state);

	if (written == -2)
		return;
	else if (written <= 0)
		gg_close_fd(f);
	else
	{
		f->twin->readoffset -= written;
		f->twin->writeoffset -= written;
		if (f->twin->readoffset != 0)
			memmove(f->twin->buffer, f->twin->buffer + written,
					f->twin->readoffset);
		if (f->twin->writeoffset == 0)
		{
			if (f->state == GG_S_CLOSING)
				gg_close_fd(f);
			else if (f->state == GG_S_WRTCMPL)
				f->expectfn(f);
			else
				gg_check_expected(f->twin);
		}
	}
}
